#<profile4xi profile = "10" />
# -------------------------------------------------------------------------------------------------
#<program4xi program = "send_handle4xi.py">
#<abstract4xi abstract = "provides the function send_handle4xi"/>
#<refer4xi reference = "reduction.py"/>
#<state4xi status = "ACK4tagging; ACK4moduletesting; NACK4systemtesting; [Rel 7.21]"/>
# -------------------------------------------------------------------------------------------------

# -------------------------------------------------------------------------------------------------
# Public API
# -------------------------------------------------------------------------------------------------
#<interface4xi interface = " 
__all__ = [ 'send_handle4xi' ] 
#"
#/>

# -------------------------------------------------------------------------------------------------
# Standard Imports                          
# -------------------------------------------------------------------------------------------------
#<standardize4xi standard = "
try:                 # system ones
    import os
    import sys
    import signal
except:
    pass
try:                 # threading ones       
    import threading                
    import _thread
except:
    pass
try:
    import _winapi
    from _winapi import WAIT_OBJECT_0, WAIT_TIMEOUT, INFINITE
except ImportError:
    if sys.platform == 'win32':
        raise
    _winapi = None
try:                 # most-common ones
    import traceback         
except:
    pass
try:                 # rare ones
    import socket          
except:
    pass 
try:                 # imported-as ones 
    pass          
except:
    pass            
#"
#/>

# -------------------------------------------------------------------------------------------------
# Multiprocessing Imports                          
# -------------------------------------------------------------------------------------------------
#<process4xi process = "
try:                             # Binary ones
    import _multiprocessing
except:
    pass
try:                             # Python ones
    from multiprocessing.reduction import send_handle
except:
    print("send_handle4xi.py: *** failed in from multiprocessing.reduction import send_handle ***")
#"
#/>

# -------------------------------------------------------------------------------------------------
# DragonXi imports                               
# -------------------------------------------------------------------------------------------------
#<dragonize4xi dragonization = "
from constant_system4xi import get_windows_system4xi
from close4xi import close4xi
from duplicate4xi import duplicate4xi
try:
    from duphandle4xi import DupHandle4Xi # new in  [Rel 7.21]
except:
    print("** [NACK] in send_handle4xi.py from duplicate4xi import DupHandle4Xi **") 
from print4xi import print4xi
from time4xi import time4xi
#" 
#/>



if sys.platform == get_windows_system4xi():
    # ---------------------------------------------------------------------------------------------
    #<function4xi function = "send-handle4xi" usage = "
    #    reduction.py                              
    #        class DupFd(object):          # however, only in Unix, not in Windows!
    #            def __init__(self, fd):
    #                new_fd = os.dup(fd)
    #            def send(conn, pid):
    #                send_handle(conn, new_fd, pid)
    #"> 
    #<note4xi note = "input parameters same but sends double handle in [Rel7.21], Windows OS version"/>
    #<state4xi status = "ACK4tagging; ACK4moduletesting; NACK4systemtesting; [Rel 7.21]"/>
    # ---------------------------------------------------------------------------------------------
    def send_handle4xi(conn, handle, destination_pid):
        '''
        created double handle from input handle and sends double handle
        '''
        
        
        # -------------------------------------------------------------------------------------    
        # (Step 1/5) - inform parameters
        # -------------------------------------------------------------------------------------         
        class2print = "send_handle4xi"
        method2print = "send_handle4xi"
    
        message2print = "{(Step 1/5a) started" + "; send-handle4xi-started: " + str(send_handle4xi)  + "; time: " + time4xi() + "}"
        print4xi(class2print, method2print, message2print) 
        # 
         
        message2print = "{(Step 1/5b) passed" + "; connection-passed-to-send-handle4xi: " + str(conn) + "; time: " + time4xi() + "}"
        print4xi(class2print, method2print, message2print)  
        # 
         
        message2print = "{(Step 1/5c) passed" + "; handle-passed-to-send-handle4xi: " + str(handle) + "; time: " + time4xi() + "}"
        print4xi(class2print, method2print, message2print)  
        # 
         
        message2print = "{(Step 1/5d) passed" + "; destination-pid-passed-to-send-handle4xi: " + str(destination_pid) + "; time: " + time4xi() + "}"
        print4xi(class2print, method2print, message2print)  
        # 
                    
        # -------------------------------------------------------------------------------------    
        # (Step 2/5) - get dublicate handler
        # -------------------------------------------------------------------------------------         
        message2print = "{(Step 2/5a) passing" + "; handle-passing-to-DupHandle4Xi-by-send-handle4xi: " + str(handle) + "; time: " + time4xi() + "}"
        print4xi(class2print, method2print, message2print)          
        
        message2print = "{(Step 2/5b) passing" + "; _winapi.DUPLICATE_SAME_ACCESS-passing-to-DupHandle4Xi-by-send-handle4xi: " + str(_winapi.DUPLICATE_SAME_ACCESS) + "; time: " + time4xi() + "}"
        print4xi(class2print, method2print, message2print)          
        
        message2print = "{(Step 2/5c) passing" + "; destination-pid-passing-to-DupHandle4Xi-by-send-handle4xi: " + str(destination_pid) + "; time: " + time4xi() + "}"
        print4xi(class2print, method2print, message2print)          
                                       
        duplicatehandler2send = DupHandle4Xi(handle, _winapi.DUPLICATE_SAME_ACCESS, destination_pid)
        
        message2print = "{(Step 2/5d) got" + "; duplicatehandler2send-got-by-send-handle4xi: " + str(duplicatehandler2send)  + "; time: " + time4xi() + "}"
        print4xi(class2print, method2print, message2print)
                
        
        
        # -------------------------------------------------------------------------------------    
        # (Step 3/5) - send with dublicate handler
        # -------------------------------------------------------------------------------------                             
      
        try:
            message2print = "{(Step 3/5a) passing" + "; duplicatehandler2send-passing-to-send-of-connection-by-send-handle4xi: " + str(duplicatehandler2send) + "; time: " + time4xi() + "}"
            print4xi(class2print, method2print, message2print)          
                    
     
            conn.send(duplicatehandler2send)
            
            message2print = "{(Step 3/5b) performed" + "; send-of-connection-performed-by-send-handle4xi: " + str(conn.send) + "; time: " + time4xi() + "}"
            print4xi(class2print, method2print, message2print)
        except:
            message2print = "{(Step 3/5z1) failed" + "; send-of-connection-failed-by-send-handle4xi: " + str(conn.send) + "; time: " + time4xi() + "}"
            print4xi(class2print, method2print, message2print)
            
            message2print = "{(Step 3/5z2) failed" + "; duplicatehandler2send-failed-by-send-handle4xi: " + str(duplicatehandler2send) + "; time: " + time4xi() + "}"
            print4xi(class2print, method2print, message2print)                             
            
        # -------------------------------------------------------------------------------------    
        # (Step 4/5) - closing not done anymore
        # -------------------------------------------------------------------------------------                    
        finally:
            pass   
        
        
        # -------------------------------------------------------------------------------------    
        # (Step 5/5) - Windows
        # ------------------------------------------------------------------------------------- 
        message2print = "{(Step 5/5) returning" + "; send-handle4xi-returning: " + time4xi() + "}" 
        print4xi(class2print, method2print, message2print)      
        
#</function4xi>
              
else:
    ACKNOWLEDGE = sys.platform == 'darwin'
    # ---------------------------------------------------------------------------------------------
    #<function4xi function = "send-handle4xi" usage = "
    #    sends handle
    #"> 
    #<state4xi status = "ACK4tagging; ACK4moduletesting; NACK4systemtesting; [Rel 7.21]"/>
    # ---------------------------------------------------------------------------------------------
    def send_handle(conn, handle, destination_pid):
        with socket.fromfd(conn.fileno(), socket.AF_UNIX, socket.SOCK_STREAM) as s:
            s.sendmsg([b'x'], [(socket.SOL_SOCKET, socket.SCM_RIGHTS,
                                struct.pack("@i", handle))])
        if ACKNOWLEDGE and conn.recv_bytes() != b'ACK':
            raise RuntimeError('did not receive acknowledgement of fd')    
                
#</function4xi>




#</program4xi>

# -------------------------------------------------------------------------------------------------
# Recorded History                             
# ------------------------------------------------------------------------------------------------- 
#<record4xi record = "
#   test_none4xi.txt
#   sense-test-none4xi.xml
#   reason-test-none4xi.xml
#   ace-test-none4xi.xml
#"
#>

# -------------------------------------------------------------------------------------------------
# Legal Notice                             
# ------------------------------------------------------------------------------------------------- 
#<protect4xi protection = "copyright"
#   owner = "Small & Smart"
#   rights = "All rights reserved"
#   state = "California"
#   country = "United States of America"
#   references = "Legal notice at www.dragonxi.com; www.yourdragonxi.com; www.yourdragonxi.org; www.dragon2xi.com/; www.dragon4xi.com/; www.dragon8xi.com/ " 
#   codes = "Codes established in California and applicable to corporations in California"
#   compliance = "The California Law"
#   court = "Supreme Court of California, 350 McAllister Street, San Francisco, CA 94102-4783"
#/>
 

#<end4xi end = "
# -------------------------------------------------------------------------------------------------
# end of send_handle4xi.py                              
# ------------------------------------------------------------------------------------------------- 
#"
#/>
